def run_abc(tokenizer, model, start_token, sep_token):
    import torch
    import pandas as pd
    from abc_lm_utilities import run_abc, eval_abc
    from transformers import AutoTokenizer, AutoModelForMaskedLM
    
    # SPECIFY HERE
    model_name = "NbAiLab/nb-bert-large"
    start_token = "[CLS] "
    sep_token = " [SEP]"

    model = AutoModelForMaskedLM.from_pretrained(model_name)
    tokenizer = AutoTokenizer.from_pretrained(model_name)
    model.eval()

    # create results df 
    df = run_abc(tokenizer, model, start_token, sep_token)
    
    # calculate difference with evaluation func
    dif = eval_abc(df)
    return dif





import json
import math
import os
import random
import sys
from pathlib import Path

import fairlearn.metrics as flm
import nltk  # only wino
import numpy as np
import pandas as pd
import progressbar
import spacy
import torch
from danlp.models import load_xlmr_coref_model
from sklearn.exceptions import UndefinedMetricWarning
from sklearn.metrics import classification_report, f1_score


def warn(*args, **kwargs):
    pass
import warnings

warnings.warn = warn

######## ABC COREFERENCE FUNCTIONS ########

def run_abc_coref(coref_model):
    ''' 
    Run the coreference model on the ABC dataset and return predictions. (Duration approx. 20 min.)
    '''
    # load stereotypically male occupation sentences
    with open(os.path.join(Path(__file__).parents[1], "data", "abc_male_sents.txt"), "r") as f:
        male_abc = f.readlines()
    # load stereotypically female occupation sentences
    with open(os.path.join(Path(__file__).parents[1], "data", "abc_fem_sents.txt"), "r") as f:
        fem_abc = f.readlines()

    all_preds = []
    for gendered_data in [fem_abc, male_abc]:  
        data_ = [line_ for line_ in gendered_data if line_ != '---\n']
        i = 0
        preds = []

        #PROGRESS BAR
        abc_bar = progressbar.ProgressBar(maxval=len(data_)).start()
        
        # run the model on the data with the predict method and save the predictions
        for idx, i in enumerate(data_): 
            line = [i.split()]
            try:
            # get preds  
                predicted_clusters = coref_model.predict(line)
            except: 
                print(line)
            preds.append(predicted_clusters)
            abc_bar.update(idx)
        abc_bar.finish()
        
        all_preds.append(preds)
    fem_preds = all_preds[0]
    male_preds = all_preds[1]

    return fem_preds, male_preds
        
def chunks(lst, n):
    """Yield successive n-sized chunks from lst."""
    for i in range(0, len(lst), n):
        yield lst[i:i + n]

def get_coref_predictions(chunk_preds):
    """Get the coreference predictions for the reflexive case and the anti-reflexive cases.
    
    Args:
        - chunk_preds: list of predictions
        - coref_output_file: the name of the file where the predictions are stored
    
    Returns:
        - fem
        - male
        - reflexive
    """
    
    ref, male, fem = [], [], []
    
    # the truth labels for the reflexive case, i.e. all 1s
    labels_ref = list(np.repeat(1, len(chunk_preds)))
    # the truth labels for the anti-reflexive cases, i.e. all 0s
    labels_anti_ref = list(np.repeat(0, len(chunk_preds)))
    
    for p in chunk_preds:
        #assert if %3 == 0
        
        ref_pred, male_pred,fem_pred = p[0], p[1], p[2]
        
        predicted_clusters = "clusters" # could be an input argument to the function to ensure that the correct key is used in case of tesitng pther models

        cluster_ref = ref_pred[predicted_clusters] 
        cluster_male = male_pred[predicted_clusters]
        cluster_fem = fem_pred[predicted_clusters] 
        clusters = [cluster_ref, cluster_male, cluster_fem]   

        for i, cluster in enumerate(clusters):
            if i == 0:
                if cluster != []:
                    ref.append(1)
                else:
                    ref.append(0)
            elif i==1:
                if cluster != []:
                    male.append(1)
                else:
                    male.append(0)
            elif i==2:
                if cluster != []:
                    fem.append(1)
                else:
                    fem.append(0)
    return ref, fem, male, labels_ref, labels_anti_ref

def get_positive_preds(chunk_preds):
    '''Get False Posivie Rates for the reflexive case and the anti-reflexive cases and print output tables.'''
    
    ref, fem, male, labels_ref, labels_anti_ref = get_coref_predictions(chunk_preds)
    labels_preds_ = zip([labels_ref,labels_anti_ref,labels_anti_ref], [ref,fem,male])                
    # the tags for the classification reports
    tags = ["reflexive", "anti_reflexive_female", "anti_reflexive_male"] 
    res = {}
    for idx, (label, pred) in enumerate(labels_preds_):
        
        if label == labels_ref:
            # get the true positive rate for the current label and prediction
            res[f"tpr_{tags[idx]}"] = flm.true_positive_rate(label, pred, pos_label=1)
        
        elif label == labels_anti_ref:
            # get the false positive rate for the current label and prediction
            res[f"fpr_{tags[idx]}"] = flm.false_positive_rate(label, pred, pos_label=1)
    print(f"""
    Total positive prediction of coreference clusters: 
    
    Female: {np.sum(fem)}
    Male: {np.sum(male)}
    Reflexive: {np.sum(ref)}
    Total predictions:  {len(fem)} {len(male)} {len(ref)}""")

    return res

def evaluate_coref_abc(fem_preds, male_preds): 
    
    for gender in ["fem", "male", "all_sents"]: #and for all 
        
        if gender == "fem":
            preds = fem_preds
        elif gender == "male":
            preds = male_preds
        elif gender == "all_sents":
            preds = fem_preds +male_preds # merge the two lists
        
        # split the predictions into chunks of 3
        chunk_preds = list(chunks(preds, 3)) #danish

        # get the false positive rates for the reflexive and anti-reflexive cases
        results = get_positive_preds(chunk_preds)

        # save the results in a dataframe
        if gender == "fem":
            df_fem = results
        elif gender == "male":
            df_male = results
        elif gender == "all_sents":
            df_all_sents = results
    
    return df_fem, df_male, df_all_sents #dicts 

def logratio(x:float ,y:float):
    return round(math.log2(x/ y), 2)

def eval_results(fem:dict, male:dict, all:dict, model_name:str):
    '''Evaluate the results of the coreference model on the ABC data.'''    
    
    data = [fem['tpr_reflexive'],
           fem['fpr_anti_reflexive_female'], 
           fem['fpr_anti_reflexive_male'], 

           male['tpr_reflexive'], 
           male['fpr_anti_reflexive_female'],
           male['fpr_anti_reflexive_male'],
           
           all['tpr_reflexive'],
           all['fpr_anti_reflexive_female'],
           all['fpr_anti_reflexive_male']
           ]
    
    mean_refl = np.mean([fem['tpr_reflexive'], male['tpr_reflexive']])
    mean_fem = np.mean([fem['fpr_anti_reflexive_female'], male['fpr_anti_reflexive_female']])
    mean_male = np.mean([fem['fpr_anti_reflexive_male'], male['fpr_anti_reflexive_male']])
    
    gender_effect_size = logratio(mean_male, mean_fem)

    data = [[round(data_,2) for data_ in data]]
    
    simlpe_data = {
        f'Simple Output for {model_name}': ['', 'ABC', ''],
        'Gender Effect Size': ['', gender_effect_size, ''],
        'Pronoun': ['Female', 'Male', 'Reflexive'],
        'Mean Rate of Detected Clusters': [mean_fem, mean_male, mean_refl]
    }

    simlpe_data = pd.DataFrame(simlpe_data).T

    # Set the first row as the column names
    simlpe_data.columns = simlpe_data.iloc[0]

    # Drop the first row
    simple_df = simlpe_data.iloc[1:]

    detailed_data = {
        f'Detailed Output for {model_name}': ['', '', 'ABC', '', '', '' ],
        'Gender Effect Size': ['', '', gender_effect_size, '', '', ''],
        'Pronoun': ['', 'Female', '', 'Male', '', 'Reflexive'],
        'Mean Rate of Detected Clusters': ['', mean_fem, '', mean_male, '', mean_refl],
        'Stereotypical Occupation': ['Female', 'Male', 'Female', 'Male', 'Female', 'Male'],
        'Rate of Detected Clusters': [fem['fpr_anti_reflexive_female'], male['fpr_anti_reflexive_female'], fem['fpr_anti_reflexive_male'], male['fpr_anti_reflexive_male'], fem['tpr_reflexive'], male['tpr_reflexive']]
        
    }

    detailed_data = pd.DataFrame(detailed_data).T

    # Set the first row as the column names
    detailed_data.columns = detailed_data.iloc[0]

    # Drop the first row
    detailed_df = detailed_data.iloc[1:]
    results = [simple_df, detailed_df]

    return results

######## WINOBIAS COREFERENCE ########

def load_texts(filepath): # NB same fn used to load wino data for lm task - no hopefully not, because it shuffles lines??       
    with open(filepath) as file:
        text = file.read().splitlines()
        lines = [line.strip() for line in text]
        random.shuffle(lines)
    return lines



def remove_sq_br(tokens):
    """Remove square brackets from tokens

    Arguments:
        tokens {[type]} -- [tokenized line]

    Returns:
        [list] -- [tokenized line without square brackets]
    """    
    #input tokens to remove '[]' 
    return [[token for token in tokens if token != '[' and token != ']']]

def remove_suffix(path, suffix) -> str:
    """
    Remove the suffix from the path if file name without extension ends with it

    Parameters
    ----------
    path
        name with prefix
    suffix
        suffix to remove

    Returns
    -------
    name_without_suffix
        name with removed suffix

    """
    if not suffix:
        return path
    directory, basename = os.path.split(path)
    basename, ext = os.path.splitext(basename)
    if basename.endswith(suffix):
        word_split = basename.rsplit(suffix, 1)
        basename = "".join(word_split[:-1])
        path = os.path.join(directory, basename + ext)
    return path 

def idx_occ_pron(tokens):
    """Get indicies of correct cluster and incorrect cluster as well as pronoun index. 
    Correct cluster: Indicies of pronoun and the occupation is refers to. 
    Incorrect cluster: Indicides of pronoun and the occupation is *does not* refer to.  

    Only used for DaNLP coref model
    """    
    #define occupations
    occupations, _ = load_occs(female = True, male=True)

    #define pronouns
    pronouns = ['hans', 'hendes', 'han', 'hun', 'ham', 'hende']

    #define square brackets 
    ## why is there only one square bracket?
    square_brackets = ['[']

    #empty lists
    occ_idx, sq_idx, prons_idx = [], [], []

    #find idx of pronouns 
    prons_idx = [[tokens.index(i) for i in pronouns if i in tokens][0] -3][0]

    #find correct referenced occupation in string
    sq_idx = [tokens.index(i) for i in square_brackets if i in tokens][0]

    #remove square brackets
    tokens = remove_sq_br(tokens)[0]
    
    #Find idx of occupations
    occ_idx = [tokens.index(token) for token in tokens if token in occupations]

    #find the incorrect referenced occupations in string
    occ_idx.remove(sq_idx)

    #save correct and incorrect answer
    correct_cluster = [sq_idx, prons_idx]
    incorrect_cluster = [occ_idx[0], prons_idx]
    
    return [correct_cluster, incorrect_cluster], prons_idx

def evaluate_model(labels, predictions):#, filename):
    """Create and save classification report
    Args: 
        labels: labels 
        predictions: model predictions
    Returns:
        clf_report: classification report
    """
    
    # create df for storing metrics
    df = pd.DataFrame(classification_report(labels,predictions,output_dict = True)).round(decimals=2)
        
    return df


def load_occs(female=False,male=False):
    """Load occupations

    Keyword Arguments:
        female {bool} -- [Whether to include occupations that defined as stereotypically female] (default: {False})
        male {bool} -- [Whether to include occupations that defined as stereotypically male] (default: {False})

    Returns:
        [Two lists] -- [One list with all occupations, one list with occupations excluding the possessive case]
    """    
    all_occupations, occupations_no_poss, occupations_poss = [], [], []
    
    # could be loaded from file
    if male:
        occupations_male = ['chaufføren', 'supervisoren', 'viceværten', 'kokken', 'flyttemanden', 
        'bygningsarbejderen', 'lederen', 'udvikleren', 'tømreren', 'bestyreren', 
        'advokaten', 'landmanden', 'sælgeren', 'lægen', 'vagten', 
        'analytikeren', 'mekanikeren', 'direktøren']
        occupations_male_poss = [occ + 's' for occ in occupations_male] # possessive case

    # could be loaded from file
    if female:
        occupations_female = ['ekspedienten', 'læreren','sygeplejersken','assistenten','sekretæren',
        'revisoren','rengøringsassistenten','receptionisten','kontorassistenten','rådgiveren',
        'designeren','frisøren','skribenten','husholdersken','bageren',
        'bogholderen','redaktøren','bibliotekaren','syersken']
        occupations_female_poss = [occ + 's' for occ in occupations_female] # possessive case

    if male and not female: 
        # occ's without pos and with pos
        occupations_no_poss, occupations_poss = occupations_male, occupations_male_poss

    if female and not male: 
        # occ's without pos and with pos
        occupations_no_poss, occupations_poss = occupations_female, occupations_female_poss

    if male and female:
        # list with occ's without possessive case
        occupations_no_poss = occupations_male + occupations_female
        # list with occ's with possessive case
        occupations_poss = occupations_male_poss + occupations_female_poss

    # list with all occupations
    all_occupations = occupations_no_poss + occupations_poss

    return all_occupations, occupations_no_poss

def run_winobias_coref(coref_model, nlp):
    """Run winobias coref experiment
    Args:
        coref_model_name: name of coref model
    Returns:
        None
    """
    
    # data paths
    inpath_pro = os.path.join(Path(__file__).parents[1],"data","DaWinoBias_pro_stereotyped_evalda.txt")
    inpath_anti = os.path.join(Path(__file__).parents[1],"data","DaWinoBias_anti_stereotyped_evalda.txt")

    # load data - nb astrid updates this fn 
    anti_lines = load_texts(inpath_anti)
    pro_lines = load_texts(inpath_pro)

    #define occupations
    occupations_male, _ = load_occs(male=True)
    occupations_female, _ = load_occs(female=True)

    occupations = occupations_male + occupations_female

    # go through each condition
    for condition in ['anti_stereotypical', 'pro_stereotypical']:
        #print(f"Condition: {condition}")
        if condition == 'anti_stereotypical':  
            lines = anti_lines 
        elif condition == 'pro_stereotypical':  
            lines = pro_lines

        # prediction results: [successful preds, unsuccessful preds, failed preds]
        pred_res = [0,0,0]
        labels, preds = [], []
        labels_occ, preds_occ = [], []
        labels_stereotype, preds_stereotype = [], []
        
        #PROGESS BAR
        bar = progressbar.ProgressBar(maxval=len(lines)).start()

        # Look through sentences
        for idx, line in enumerate(lines): 
            # read line as spacy doc
            line = nlp(line)

            # tokenize and lowercase
            tokens = []
            for token in line:
                tokens.append(token.text.lower())

            # get correct coref and incorrect coref to compare with predictions
            coref_res,_ = idx_occ_pron(tokens)
            
            # remove square brackets aka the solution before feeding into coref model
            tokens = remove_sq_br(tokens)
            
            # apply coreference resolution to the document and get a list of predicted coref clusters (not idx)
            clusters = coref_model.predict_clusters(tokens)

            # get token indices from predicted cluster
            
            #if len(clusters) == 0:
            if clusters == []:
                cluster_idx = [-1]
            else:
                cluster_idx = [i[1] for i in clusters[0]]
            
            # compare predicted clusters with correct res
            if cluster_idx == coref_res[0]:
                pred_res[0] += 1
            elif cluster_idx == coref_res[1]:
                pred_res[1] += 1
            else: 
                pred_res[2] += 1

            # labels 
            labels.append(coref_res[0][0])
            
            #predictions
            if len(cluster_idx)>2 or len(cluster_idx)<1: 
                preds.append(-1)
            elif len(cluster_idx)==2:
                preds.append(cluster_idx[0])
            
            #occupation labels
            labels_occ.append(tokens[0][coref_res[0][0]])
            preds_occ.append(tokens[0][cluster_idx[0]])

            #remove possisive occupations 
            labels_occ = [remove_suffix(label, 's') for label in labels_occ]
            preds_occ = [remove_suffix(pred, 's') for pred in preds_occ]

            #group occupations
            for label, pred in zip(labels_occ, preds_occ):
                if label in occupations_female and pred in occupations: 
                    labels_stereotype.append('stereotypical female')
                elif label in occupations_male  and pred in occupations:
                    labels_stereotype.append('stereotypical male')

            for pred in preds_occ:
                if pred in occupations_female: 
                    preds_stereotype.append('stereotypical female')
                elif pred in occupations_male:
                    preds_stereotype.append('stereotypical male')

            # update bar
            bar.update(idx)
        bar.finish()

        #remove invalid predictions
        labels_occ = [labels_occ for labels_occ, preds_occ in zip(labels_occ, preds_occ) if preds_occ in occupations]
        preds_occ = [preds_occ for preds_occ in preds_occ if preds_occ in occupations]

        # get results in table
        # results per occupations, save file
        evaluate_model(labels_occ, preds_occ) #, filename = f'output/winobias_{coref_model_name}_{condition}_occupations') #switch to clf function
        # results per stereotype, save file
        wino_res_ = evaluate_model(labels_stereotype, preds_stereotype) #, filename = f'output/wino_bias{coref_model_name}_{condition}_stereotypes') #switch to clf function

        if condition == 'anti_stereotypical':
            anti_res = wino_res_
        elif condition == 'pro_stereotypical':
            pro_res = wino_res_
    return anti_res, pro_res

def evaluate_coref_winobias(anti_res, pro_res, model_name):
    """Evaluate winobias coref experiment
    Args:
        anti_res: results for anti-stereotypical condition
        pro_res: results for pro-stereotypical condition
        
    """
    # Gender Effect Size calculation
    anti_acc = anti_res["accuracy"].values[0]
    pro_acc = pro_res["accuracy"].values[0]
    gender_effect_size = logratio(pro_acc, anti_acc)

    anti_f1_fem_pron = anti_res["stereotypical female"].values[2]
    anti_f1_male_pron = anti_res["stereotypical male"].values[2]

    pro_f1_fem_pron = pro_res["stereotypical female"].values[2]
    pro_f1_male_pron = pro_res["stereotypical male"].values[2]

    simlpe_data = {
        f'Simple Output for {model_name}': ['', 'DaWinoBias'],
        'Gender Effect Size': ['', gender_effect_size],
        'Condition': ['Anti-stereotypical','Pro-stereotypical'],
        'Accuracy': [anti_acc, pro_acc]
        }

    simlpe_data = pd.DataFrame(simlpe_data).T

    # Set the first row as the column names
    simlpe_data.columns = simlpe_data.iloc[0]

    # Drop the first row
    simple_df = simlpe_data.iloc[1:]

    detailed_data = {
        f'Detailed Output for {model_name}': ['', 'DaWinoBias', '',''],
        'Gender Effect Size': ['', gender_effect_size, '',''],
        'Condition': ['Anti-stereotypical','', 'Pro-stereotypical', ''],
        'Accuracy': [anti_acc, '', pro_acc, ''],
        'Pronouns': ['Female', 'Male', 'Female', 'Male'],
        'F1': [anti_f1_fem_pron, anti_f1_male_pron, pro_f1_fem_pron, pro_f1_male_pron]
        }

    detailed_data = pd.DataFrame(detailed_data).T

    # Set the first row as the column names
    detailed_data.columns = detailed_data.iloc[0]

    # Drop the first row
    detailed_df = detailed_data.iloc[1:]

    results = [simple_df, detailed_df]

    return results


"""
Utility functions for running the ABC language modeling task.

Builds on the codebase developed for the following paper: 

Title: “Type B reflexivization as an unambiguous testbed for multilingual multi-task gender bias.”
Authors: González, A. V., Barrett, M., Hvingelby, R., Webster, K., & Søgaard, A.
Date: 2020
Code availability: https://github.com/anavaleriagonzalez/ABC-dataset
"""
import math
import os
import sys

import numpy as np
import pandas as pd
import torch
from tqdm import tqdm

def load_mdl(model_name):
    print(f"[INFO] Loading model {model_name} from Hugging Face.")
    from transformers import AutoModelForMaskedLM, AutoTokenizer
    model = AutoModelForMaskedLM.from_pretrained(model_name)
    tokenizer = AutoTokenizer.from_pretrained(model_name)
    model.eval()
    return model, tokenizer

def load_abc():
    from pathlib import Path 
    inpath_male = os.path.join(Path(__file__).parents[1],"data","abc_male_sents.txt")
    inpath_fem = os.path.join(Path(__file__).parents[1],"data","abc_fem_sents.txt")
    male_sents = load_sents(inpath_male)
    fem_sents = load_sents(inpath_fem)
    return male_sents, fem_sents

def load_sents(filename):
    """
    A function that loads all the reflexive sentences from the ABC dataset.
    """
    reflexive_sents = []
    with open(filename, "r") as f:
        lines = f.readlines()

        restart = 0
        for line in lines:
            if "--------------" in line: pass #is this needed?
            elif "---" in line:
                restart = 0
            else:
                if restart == 0:
                    reflexive_sents.append(line.strip())
                    restart = 1
    return reflexive_sents

def tokenize_sentence(sentence, tokenizer, start_token, sep_token): # add here if not a bert model
    """
    A function that tokenizes the reflexive sentences.
    """
    sentence = start_token+sentence+sep_token 
    tokenize_input = tokenizer.tokenize(sentence)
    return tokenize_input

def get_pron_index(sent, pronoun_list):
    """
    A function that takes the tokenized reflexive sent and
    locates index of pronoun
    replaces pronoun with male/female pronoun 
    returns 2 augmented sentences 
    """
    # get index of pronoun
    no_pron = True
    for i, token in enumerate(sent):
        if token in pronoun_list:
            pron_index = i
            no_pron = False
            break
        else:pass
    if no_pron==True: return "no pronouns to replace"
    else: return pron_index 

def get_augmented_sents(tokenize_input, idx):  
    """
    Take original sentence and replace pronoun with antirreflexive pronouns (hans/hendes) 
    """
    tokenize_mask_male = tokenize_input.copy()
    tokenize_mask_female = tokenize_input.copy()
    tokenize_mask_male[int(idx)] = "hans"
    tokenize_mask_female[int(idx)] = "hendes"
    return tokenize_mask_male, tokenize_mask_female

def create_tensors(truth, male, fem, tokenizer):
    """
    Make augmented sentences into tensors.
    """
    tensor_truth = torch.tensor([tokenizer.convert_tokens_to_ids(truth)])
    tensor_male = torch.tensor([tokenizer.convert_tokens_to_ids(male)])
    tensor_fem = torch.tensor([tokenizer.convert_tokens_to_ids(fem)])    
    return tensor_truth, tensor_male, tensor_fem

def get_predictions(tensor_truth, tensor_male, tensor_fem, model):
    """
    A function that takes the 3 sentences and returns model predictions for each of them.
    Used to compute loss.   
    """
    with torch.no_grad():
        pred_truth = model(tensor_truth)[0]
    with torch.no_grad():
        pred_male = model(tensor_male)[0]
    with torch.no_grad():
        pred_fem = model(tensor_fem)[0]
    return pred_truth, pred_male, pred_fem

def compute_loss(loss_fct, pred_truth, pred_male, pred_fem, tensor_truth):
    """
    Take model predictions and compute loss for all 3 sentences 
    by comparing to pred_truth, which is the prediction of the sentence with the correct (reflexive) pronoun.
    """
    loss_male = loss_fct(pred_male.squeeze(),tensor_truth.squeeze()).data
    loss_fem = loss_fct(pred_fem.squeeze(),tensor_truth.squeeze()).data
    loss_ref = loss_fct(pred_truth.squeeze(),tensor_truth.squeeze()).data
    loss_list = [loss_male, loss_fem, loss_ref]
    return loss_list

def score_sent(sent, loss_fct, tokenizer, model, pron_list, start_token , sep_token):
    """
    Tak a sentence with a relflexive pronoun, 
    replace pronoun with antireflexives (male/female pronoun)
    and compute loss and perplexity for all 3 sentences.
    Args:
        sent (str): A sentence with reflexive pronouns.
        loss_fct (torch.nn.CrossEntropyLoss): A loss function.
        tokenizer (BertTokenizer): A tokenizer.
        model (BertForMaskedLM): A language model.
        pron_list (list): A list of pronouns.
    Returns:
        Loss and perplexity values for sentence with reflexive, male and female pronoun. 
    """
    tokenized_refl = tokenize_sentence(sent, tokenizer, start_token, sep_token)
    index = get_pron_index(tokenized_refl, pron_list)
    tokenized_male, tokenized_fem = get_augmented_sents(tokenized_refl, index)
    tensor_truth, tensor_male, tensor_fem = create_tensors(tokenized_refl, tokenized_male, tokenized_fem, tokenizer)
    pred_truth, pred_male, pred_fem = get_predictions(tensor_truth, tensor_male, tensor_fem, model)
    loss_values = compute_loss(loss_fct, pred_truth, pred_male, pred_fem, tensor_truth)
    return str(math.exp(loss_values[0])), str(math.exp(loss_values[1])), str(math.exp(loss_values[2]))

def run_abc(reflexive_sents, condition, tokenizer, model, start_token , sep_token):
    """
    Run the ABC dataset and write results to file.
    Args:
        outpath (str): Path to output file.
        reflexive_sents (list): A list of sentences with reflexive pronouns.
        loss_fct (torch.nn.CrossEntropyLoss): A loss function.
        tokenizer (BertTokenizer): A tokenizer.
        model (BertForMaskedLM): A language model.
        pron_list (list): A list of pronouns.
    """

    # loss function for making model predictions and loss - used for getting pperplexity scores
    loss_fct = torch.nn.CrossEntropyLoss()

    pronouns_list = ["sin", "sit", "sine", "▁sin", "▁sit", "▁sine"]

    print(f"[INFO] Running test on condition: {condition} occupations.")
    # loop over sentences to compute loss and perplexity
    sents, perp_m,perp_f,perp_r = [],[],[],[]
    for idx, sent in enumerate(tqdm(reflexive_sents)):
        scores = score_sent(sent, loss_fct, tokenizer, model, pronouns_list, start_token, sep_token)
        sents.append(sent)
        perp_m.append(scores[0])
        perp_f.append(scores[1])
        perp_r.append(scores[2])


    df = pd.DataFrame({"sent":sents, 
                      "perplexity_male":perp_m,
                      "perplexity_female":perp_f,
                      "perplexity_reflexive":perp_r})
    
    df["perplexity_male"] = df["perplexity_male"].astype(float)
    df["perplexity_female"] = df["perplexity_female"].astype(float)
    df["perplexity_reflexive"] = df["perplexity_reflexive"].astype(float)

    return df


### EVALUTATION
def logratio(x:float,y:float):
    return round(math.log2(x/ y), 2)

def calc_median_iqr(df, l_ratio=False):
    # return median value for male/female pronoun with IQR
    # fem 
    median_fem = str(df.relative_female.median().round(2))
    iqr_fem = str(tuple(np.percentile(df.relative_female, [25,75]).round(2)))
    
    # male 
    median_male = str(df.relative_male.median().round(2))
    iqr_male = str(tuple(np.percentile(df.relative_male, [25,75]).round(2)))
    if l_ratio == True:
        ratio = logratio(float(median_fem), float(median_male)) # female over male here! because high perplexity = bad 
        return str(median_fem+" "+iqr_fem), str(median_male+" "+iqr_male), ratio
    elif l_ratio == False:
        return str(median_fem+" "+iqr_fem), str(median_male+" "+iqr_male)

def get_condensed(all_occ_FEM_PRON, all_occ_MALE_PRON, lratio, model_name):
    data = {
    f'Simple Output: {model_name}': ['', 'ABC', '', ''],
    'Gender Effect Size': ['', lratio,'', ''],
    'Pronoun': ['Female', '', '', 'Male'],
    'Perplexity Median (IQR) Relative scores': ['', all_occ_FEM_PRON, '', all_occ_MALE_PRON]}
    df = pd.DataFrame(data).T
    df.columns = df.iloc[0]
    df = df.iloc[1:]
    return df

def get_detailed(all_occ_FEM_PRON, all_occ_MALE_PRON, lratio, fem_occ_FEM_PRON, fem_occ_MALE_PRON,male_occ_FEM_PRON, male_occ_MALE_PRON, model_name):
    data = {f'Detailed Output for: {model_name}': ['', 'ABC', '', ''],
        'Gender Effect Size': ['', lratio,'', ''],
        'Pronoun': ['Female', '', 'Male', ''],
        'Perplexity Median (IQR) Relative scores': [all_occ_FEM_PRON, '', all_occ_MALE_PRON, ''],
        'Stereotypical Occupation': ['Female', 'Male', 'Female', 'Male'],
        'Perplexity Median (IQR) Relative scores ': [fem_occ_FEM_PRON, male_occ_FEM_PRON, fem_occ_MALE_PRON, male_occ_MALE_PRON]}
    df = pd.DataFrame(data).T
    df.columns = df.iloc[0]
    df = df.iloc[1:]
    return df

def get_output(df_FEM_occ, df_MALE_occ, model_name):
    # calculate relative perplexities 
    df_FEM_occ["relative_female"] = df_FEM_occ["perplexity_female"] - df_FEM_occ["perplexity_reflexive"]
    df_FEM_occ["relative_male"] = df_FEM_occ["perplexity_male"] - df_FEM_occ["perplexity_reflexive"]
    df_MALE_occ["relative_female"] = df_MALE_occ["perplexity_female"] - df_MALE_occ["perplexity_reflexive"]
    df_MALE_occ["relative_male"] = df_MALE_occ["perplexity_male"] - df_MALE_occ["perplexity_reflexive"]

    # create combined df  
    ALL_occ = pd.concat([df_FEM_occ, df_MALE_occ], ignore_index=True)

    # calculate median and IQR's 
    fem_occ_FEM_PRON, fem_occ_MALE_PRON = calc_median_iqr(df_FEM_occ)
    male_occ_FEM_PRON, male_occ_MALE_PRON = calc_median_iqr(df_MALE_occ)
    all_occ_FEM_PRON, all_occ_MALE_PRON, log_ratio = calc_median_iqr(ALL_occ, l_ratio=True)

    # get condensed output
    out_condensed = get_condensed(all_occ_FEM_PRON, all_occ_MALE_PRON, log_ratio, model_name)

    # get detailed output
    out_detailed = get_detailed(all_occ_FEM_PRON, all_occ_MALE_PRON, log_ratio, fem_occ_FEM_PRON, fem_occ_MALE_PRON,male_occ_FEM_PRON, male_occ_MALE_PRON, model_name)

    return [out_condensed, out_detailed]



""" 
Utility functions for running the NER task.

Builds on early versions of the codebase developed for the following publication: 

Title: “Detecting intersectionality in NER models: A data-driven approach.”
Authors: Lassen, I. M., Almasi, M., Enevoldsen, K., & Kristensen-mclachlan, R. 
Date: 2023
Code availability: https://github.com/centre-for-humanities-computing/Danish-NER-bias
"""

import math
import os

import pandas as pd
import spacy
from dacy.score import score

def load_mdl(model_name):
    """Loading pre-trained model using DaCy or from Hugging Face via spacy_wrap 

    Args:
        model_name (str): name of model in DaCy or on Hugging Face 

    Raises:
        ValueError: raise error if HF model cannot be loaded from Hugging Face
        ValueError: raise error if model type not supported in package 

    Returns:
        model: NER model  
    """    
    import dacy
    if model_name in dacy.models():
        print(f"[INFO] Loading model {model_name} with DaCy.")
        import dacy

        #model = dacy.load("da_dacy_large_trf-0.2.0")
        model = dacy.load(model_name)
    elif model_name not in dacy.models():
        print(f"[INFO] Loading model {model_name} from Hugging Face with spaCy-wrap.")
        try:
            #import spacy
            import spacy_wrap
            model = spacy.blank("da")
            # config for loading 
            config = {"model": {"name": model_name}, 
                        "predictions_to": ["ents"]} 
            model.add_pipe("token_classification_transformer", config=config)
        except OSError:
            raise ValueError("Cannot load model from Hugging Face. Please specify the full name of the model i.e. 'name/model-name' ")     
    else:
        raise ValueError("Cannot load model. This package only supports models from DaCy and Hugging Face.")
    return model

def eval_model_augmentation(mdl, model_name, n, augmenters, dataset):
    """Compute NER performance on the DaNe test set using DaCy's score function

    Args:
        mdl (model): NER model  
        model_name (str): model name
        n (int): number of repetitions to run 
        augmenters (generator): dictionary with name augmentaions to test 
        dataset (corpus): the DaNe test set 

    Returns:
        list: list with two df's: output condensed and output detailed  
    """    
    # loop over all models in model_dict 
    scores = []
    i = 0
    for aug, nam, k in augmenters:
        print(f"\t [INFO] Running augmenter: {nam} with {k} repetitions.")
        scores_ = score(corpus=dataset, apply_fn=mdl, augmenters=aug, k=k)
        scores_["model"] = model_name
        scores_["augmenter"] = nam
        scores_["i"] = i
        scores.append(scores_)
        i += 1
    scores_df = pd.concat(scores)

    # summarise scores over augmentations, compute gender effect
    scores_summed, l_ratio = get_means_sds(scores_df)

    # compute simple and detailed output
    out_detailed = eval_ner_detailed(scores_summed, l_ratio, model_name)

    out_simple = eval_ner_simple(scores_summed, l_ratio, model_name)

    return [out_simple, out_detailed]

def logratio(x:float ,y:float):
    return round(math.log2(x/ y), 2)

def get_means_sds(df):
    '''
    Compute mean F1 and SD for each augmenter of a single model.

    input:
     - df with model performance of a single model. 

    output: 
    -  
    '''
    print(df.head())
    df["f1"] = df["ents_excl_MISC_ents_f"] #*100
    df = df[df["augmenter"].isin(["Majority female names", "Majority male names", "Minority male names", "Minority female names"])]
    augs, means, sds = [],[],[]
    for aug in set(df["augmenter"]):
        augs.append(aug)
        sub = df[df["augmenter"]==aug]
        means.append(round(sub["f1"].mean(),4))
        sds.append(round(sub["f1"].std(),2))
    # create df
    out_df = pd.DataFrame({"Group": augs, "F1": means, "SD": sds})
    # create df with mean and SD in one column
    combinations = [str(str(a)+" ("+str(b)+")") for (a,b) in zip(means,sds)]
    out_df_condensed = pd.DataFrame({"Group": augs, "F1 (SD)": combinations})

    # Compute log-ratio
    f = out_df[out_df["Group"]=="Majority female names"].iloc[0,1]
    m = out_df[out_df["Group"]=="Majority male names"].iloc[0,1]
    ratio = logratio(m,f)  

    return out_df_condensed, ratio


def eval_ner_simple(df, l_ratio, model_name):
    data = {
        f'Simple Output {model_name}': ['', 'Augmented DaNe', ''],
        'Gender Effect Size': ['', l_ratio,''],
        'Name Augmentation': ['Female', '', 'Male'],
        'Macro F1 (SD)': [df[df["Group"]=="Majority female names"].iloc[0,1], '', 
                     df[df["Group"]=="Majority male names"].iloc[0,1]],
    }
    df_out = pd.DataFrame(data).T
    df_out.columns = df_out.iloc[0]
    df_out = df_out.iloc[1:]
    return df_out

def eval_ner_detailed(df, l_ratio, model_name):
    data = {
        f'Detailed Output for {model_name}': ['', 'Augmented DaNe', '', ''],
        'Gender Effect Size': ['', l_ratio,'', ''],
        'Name Augmentation': ['Female', '', 'Male', ''],
        'Macro F1 (SD)': [df[df["Group"]=="Majority female names"].iloc[0,1],'', df[df["Group"]=="Majority male names"].iloc[0,1],''],
        'Name Augmentation ': ['Majority Female', 'Minority Female', 'Majority Male', 'Minority Male'],
        'Macro F1 (SD) ': [df[df["Group"]=="Majority female names"].iloc[0,1], 
                      df[df["Group"]=="Minority female names"].iloc[0,1], 
                      df[df["Group"]=="Majority male names"].iloc[0,1], 
                      df[df["Group"]=="Minority male names"].iloc[0,1]]
    }
    out_df = pd.DataFrame(data).T
    out_df.columns = out_df.iloc[0]
    out_df = out_df.iloc[1:]
    return out_df




""" 
Utility functions for creating the dictionaries with name augmentations used to run the NER task.

Builds on early versions of the codebase developed for the following publication: 

Title: “Detecting intersectionality in NER models: A data-driven approach.”
Authors: Lassen, I. M., Almasi, M., Enevoldsen, K., & Kristensen-mclachlan, R. 
Date: 2023
Code availability: https://github.com/centre-for-humanities-computing/Danish-NER-bias
"""
import augmenty
from .process_names import (f_name_dict, m_name_dict, muslim_f_dict,
                            muslim_m_dict)

# define pattern of augmentation
patterns = [["first_name"], ["first_name", "last_name"],
            ["first_name", "last_name", "last_name"]]

# define person tag for augmenters 
person_tag = "PER" 

# define all augmenters 
f_aug = augmenty.load(
    "per_replace_v1", 
    patterns = patterns, 
    names = f_name_dict, 
    level = 1, 
    person_tag = person_tag, 
    replace_consistency = True
    )

m_aug = augmenty.load(
    "per_replace_v1", 
    patterns = patterns, 
    names = m_name_dict, 
    level = 1, 
    person_tag = person_tag, 
    replace_consistency = True
    )

muslim_f_aug = augmenty.load(
    "per_replace_v1", 
    patterns = patterns, 
    names = muslim_f_dict, 
    level = 1, 
    person_tag = person_tag, 
    replace_consistency = True
    )

muslim_m_aug = augmenty.load(
    "per_replace_v1", 
    patterns = patterns, 
    names = muslim_m_dict, 
    level = 1, 
    person_tag = person_tag, 
    replace_consistency = True
    )



""" 
Utility functions for creating the names lists used to run the NER task.

Builds on early versions of the codebase developed for the following publication: 

Title: “Detecting intersectionality in NER models: A data-driven approach.”
Authors: Lassen, I. M., Almasi, M., Enevoldsen, K., & Kristensen-mclachlan, R. 
Date: 2023
Code availability: https://github.com/centre-for-humanities-computing/Danish-NER-bias
"""
import os
from pathlib import Path

import pandas as pd
from dacy.datasets import load_names, muslim_names

def remove_duplicates(all_names, names_to_filter_away):
    all_names = [name for name in all_names if name not in names_to_filter_away]
    return all_names

# path to names files 
data_folder = os.path.join(Path(__file__).parents[1], "data","name_aug_csv_files") 

### Define majority/danish names

# get last names 
last_names_2023 = pd.read_csv(os.path.join(data_folder, "last_names_2023.csv"))
last_names_2023["Navn"] = last_names_2023["Navn"].str.title() # capitalize
last_names_2023 = list(last_names_2023["Navn"])[:500] # subset to only 500 to match 500 first names

# men and women first names
men_2023 = pd.read_csv(os.path.join(data_folder, "first_names_2023_men.csv"))
women_2023 = pd.read_csv(os.path.join(data_folder, "first_names_2023_women.csv"))

# capitalize
men_2023["Navn"] = men_2023["Navn"].str.title()
women_2023["Navn"] = women_2023["Navn"].str.title()

# subset names to 500 
men_2023 = list(men_2023["Navn"])[:500]
women_2023 = list(women_2023["Navn"])[:500]

# create dictionaries 
m_name_dict = {'first_name':men_2023, 'last_name':last_names_2023}
f_name_dict = {'first_name':women_2023, 'last_name':last_names_2023}

### Define muslim/minority names 
muslim_name_dict = muslim_names()
muslim_m_dict = load_names(ethnicity="muslim", gender="male", min_prop_gender=0.5)
muslim_f_dict = load_names(ethnicity="muslim", gender="female", min_prop_gender=0.5)

### Remove overlaps 
# read in annotated
overlaps = pd.read_csv(os.path.join(data_folder, "overlapping_names.csv"))

# create muslim/minority only list and majority/danish only list 
muslim_only=list(overlaps.query("origin=='M'")["duplicates"])
danish_only=list(overlaps.query("origin=='DK'")["duplicates"])

# majority/danish seperate genders into seperate dicts
f_name_dict["first_name"] = remove_duplicates(f_name_dict["first_name"], muslim_only)
m_name_dict["first_name"] = remove_duplicates(m_name_dict["first_name"], muslim_only)

# muslim/minority seperate genders into seperate dicts
muslim_f_dict["first_name"] = remove_duplicates(muslim_f_dict["first_name"], danish_only)
muslim_m_dict["first_name"] = remove_duplicates(muslim_m_dict["first_name"], danish_only)




"""
Utility functions for running the DaWinobias language modeling task.

Builds on the codebase developed for the following project: 

Title: “DaWinoBias: Assessing Occupational Gender Stereotypes in Danish NLP Models”
Authors: Koppelgaard, K., Brødbæk, S. K.
Date: 2021
Code availability: https://github.com/NLP-exam/DaWinoBias
"""
import math
import os
import random
from collections import Counter
from pathlib import Path

import pandas as pd
import spacy
import torch
from sklearn.metrics import classification_report
from tqdm import tqdm


def load_texts(filepath):
    """load DaWinoBias texts - and shuffle data 
    """    
    with open(filepath) as file:
        text = file.read().splitlines()
    lines = [line.strip() for line in text]
    random.shuffle(lines)
    return lines

def remove_sq_br(tokens):  
    #input tokens to remove '[]' 
    return [[token for token in tokens if token != '[' and token != ']']]

def idx_pron(tokens):
    """Get index of pronoun in sentence
    """    
    #define pronouns
    pronouns = ['hans', 'hendes', 'han', 'hun', 'ham', 'hende','▁hun', '▁hendes', '▁hende', '▁han', '▁hans', '▁ham']
    # find idx of pronouns 
    prons_idx = [tokens.index(i) for i in pronouns if i in tokens][0]
    return prons_idx

def predict_masked(lines, condition, nlp, tokenizer, mask_token, model_name): 
    labels, preds = [], []

    print(f"[INFO] Running test on condition: {condition} sentences.")
    for idx, line in enumerate(tqdm(lines)):
        # tokenize and lowercase
        line_tokenized = tokenizer(line)
        tokens = [token.text.lower() for token in line_tokenized]
        
        # remove square brackets from tokens
        tokens = remove_sq_br(tokens)[0]

        # find index of pronoun
        prons_idx = idx_pron(tokens)

        #save correct pronoun
        correct_pronoun = tokens[prons_idx]
        
        # MASK pronouns
        tokens[prons_idx] = mask_token
                 
        # join sentence into one string 
        if mask_token == '<mask>' and model_name == "vesteinn/DanskBert": 
            sentence = ' '.join(tokens).replace(" <mask>", "<mask>")
        else: 
            sentence = ' '.join(tokens)
        
        # compute fill-mask prediction
        pred = nlp(sentence)[0]['token_str']
        
        # save labels and predictions
        labels.append(correct_pronoun)
        preds.append(pred)
        
    return labels, preds

def group_pronouns(pronouns): 
    '''group labels in female and male pronouns'''

    #define groups of pronouns
    female_pronouns = ['hun', 'hendes', 'hende', '▁hun', '▁hendes', '▁hende']
    male_pronouns = ['han', 'hans', 'ham', '▁han', '▁hans', '▁ham']

    #initialize lists
    group_pronoun= []

    #group labels
    for pronoun in pronouns: 
        if pronoun in female_pronouns: 
            group_pronoun.append('hun/hendes')
        elif pronoun in male_pronouns: 
            group_pronoun.append('han/hans')
        else:
            group_pronoun.append('UNK')
    return group_pronoun

def evaluate_model(labels, predictions): #, filename, model_name):
    """Create and save classification report
    Args: 
        labels: labels 
        predictions: model predictions
    Returns:
        clf_report: classification report in pandas df format
    """
    # create df for storing metrics
    df = pd.DataFrame(classification_report(labels,predictions,output_dict = True, zero_division=1)).round(decimals=2)
    return df 


def run_winobias(tokenizer, nlp, mask_token, model_name): 
    # data paths
    inpath_pro = os.path.join(Path(__file__).parents[1],"data","DaWinoBias_pro_stereotyped_evalda.txt")
    inpath_anti = os.path.join(Path(__file__).parents[1],"data","DaWinoBias_anti_stereotyped_evalda.txt")

    # load data
    anti_sents = load_texts(inpath_anti)
    pro_sents = load_texts(inpath_pro)

    # mask and predict pronoun
    anti_labels_, anti_preds_ = predict_masked(lines = anti_sents, condition = "anti-stereotypical", nlp = nlp, tokenizer = tokenizer, mask_token = mask_token, model_name = model_name)
    pro_labels_, pro_preds_ = predict_masked(lines = pro_sents, condition = "pro-stereotypical", nlp = nlp, tokenizer = tokenizer, mask_token = mask_token, model_name = model_name)

    # group pronouns into male/female category 
    anti_labels, anti_preds = group_pronouns(anti_labels_),group_pronouns(anti_preds_) 
    pro_labels, pro_preds = group_pronouns(pro_labels_),group_pronouns(pro_preds_) 
    print( "[INFO] Raw predictions for model:", model_name)
    
    print("[INFO] Number of female pron. predictions:", len([i for i in anti_preds if i =='hun/hendes']))
    print("[INFO] Number of male pron. predictions:", len([i for i in anti_preds if i =='han/hans']))
    print("[INFO] Number of 'UNKNOWN' predictions:", len([i for i in anti_preds if i =='UNK']))
    
    # evaluate performance
    clf_rep_anti = evaluate_model(anti_labels, anti_preds)
    clf_rep_pro = evaluate_model(pro_labels, pro_preds)
    
    return clf_rep_anti, clf_rep_pro

def logratio(x:float ,y:float):
    return round(math.log2(x/ y), 2)

def evaluate_lm_winobias(anti_res, pro_res, model_name):
    """Evaluate winobias coref experiment
    Args:
        anti_res: results for anti-stereotypical condition
        pro_res: results for pro-stereotypical condition
        
    """
    # Gender Effect Size calculation
    anti_acc = anti_res["accuracy"].values[0]
    pro_acc = pro_res["accuracy"].values[0]
    gender_effect_size = logratio(pro_acc, anti_acc)

    anti_f1_fem_pron = anti_res["hun/hendes"].values[2]
    anti_f1_male_pron = anti_res["han/hans"].values[2]

    pro_f1_fem_pron = pro_res["hun/hendes"].values[2]
    pro_f1_male_pron = pro_res["han/hans"].values[2]

    simlpe_data = {
        f'Simple Output for {model_name}': ['', 'DaWinoBias'],
        'Gender Effect Size': ['', gender_effect_size],
        'Condition': ['Anti-stereotypical','Pro-stereotypical'],
        'Accuracy': [anti_acc, pro_acc]
        }

    simlpe_data = pd.DataFrame(simlpe_data).T

    # Set the first row as the column names
    simlpe_data.columns = simlpe_data.iloc[0]

    # Drop the first row
    simple_df = simlpe_data.iloc[1:]
    
    detailed_data = {
        f'Detailed Output for {model_name}': ['', 'DaWinoBias', '',''],
        'Gender Effect Size': ['', gender_effect_size, '',''],
        'Condition': ['Anti-stereotypical','', 'Pro-stereotypical', ''],
        'Accuracy': [anti_acc, '', pro_acc, ''],
        'Pronouns': ['Female', 'Male', 'Female', 'Male'],
        'F1': [anti_f1_fem_pron, anti_f1_male_pron, pro_f1_fem_pron, pro_f1_male_pron]
        }

    detailed_data = pd.DataFrame(detailed_data).T

    # Set the first row as the column names
    detailed_data.columns = detailed_data.iloc[0]

    # Drop the first row
    detailed_df = detailed_data.iloc[1:]

    results = [simple_df, detailed_df]

    return results



"""
Utility functions for running the DaWinobias coreference task.

Builds on the codebase developed for the following project: 

Title: “DaWinoBias: Assessing Occupational Gender Stereotypes in Danish NLP Models”
Authors: Koppelgaard, K., Brødbæk, S. K.
Date: 2021
Code availability: https://github.com/NLP-exam/DaWinoBias
"""
import json
import math
import os
import random
import sys
from pathlib import Path

import fairlearn.metrics as flm
import nltk  # only wino
import numpy as np
import pandas as pd
import spacy
import torch
from sklearn.exceptions import UndefinedMetricWarning
from sklearn.metrics import classification_report, f1_score

def load_texts(filepath):  
    '''
    load DaWinoBias texts - and shuffle data 
    '''    
    with open(filepath) as file:
        text = file.read().splitlines()
        lines = [line.strip() for line in text]
        random.shuffle(lines)
    return lines

def remove_sq_br(tokens):
    """Remove square brackets from tokens

    Arguments:
        tokens {[type]} -- [tokenized line]

    Returns:
        [list] -- [tokenized line without square brackets]
    """    
    #input tokens to remove '[]' 
    return [[token for token in tokens if token != '[' and token != ']']]

def remove_suffix(path, suffix) -> str:
    """
    Remove the suffix from the path if file name without extension ends with it

    Parameters
    ----------
    path
        name with prefix
    suffix
        suffix to remove

    Returns
    -------
    name_without_suffix
        name with removed suffix

    """
    if not suffix:
        return path
    directory, basename = os.path.split(path)
    basename, ext = os.path.splitext(basename)
    if basename.endswith(suffix):
        word_split = basename.rsplit(suffix, 1)
        basename = "".join(word_split[:-1])
        path = os.path.join(directory, basename + ext)
    return path 

def idx_occ_pron(tokens):
    """Get indicies of correct cluster and incorrect cluster as well as pronoun index. 
    Correct cluster: Indicies of pronoun and the occupation is refers to. 
    Incorrect cluster: Indicides of pronoun and the occupation is *does not* refer to.  

    Only used for DaNLP coref model
    """    
    #define occupations
    occupations, _ = load_occs(female = True, male=True)

    #define pronouns
    pronouns = ['hans', 'hendes', 'han', 'hun', 'ham', 'hende']

    #define square brackets 
    ## why is there only one square bracket?
    square_brackets = ['[']

    #empty lists
    occ_idx, sq_idx, prons_idx = [], [], []

    #find idx of pronouns 
    prons_idx = [[tokens.index(i) for i in pronouns if i in tokens][0] -3][0]

    #find correct referenced occupation in string
    sq_idx = [tokens.index(i) for i in square_brackets if i in tokens][0]

    #remove square brackets
    tokens = remove_sq_br(tokens)[0]
    
    #Find idx of occupations
    occ_idx = [tokens.index(token) for token in tokens if token in occupations]

    #find the incorrect referenced occupations in string
    occ_idx.remove(sq_idx)

    #save correct and incorrect answer
    correct_cluster = [sq_idx, prons_idx]
    incorrect_cluster = [occ_idx[0], prons_idx]
    
    return [correct_cluster, incorrect_cluster], prons_idx

def evaluate_model(labels, predictions): 
    """Create and save classification report
    Args: 
        labels: labels 
        predictions: model predictions
    Returns:
        clf_report: classification report
    """
    
    # create df for storing metrics
    df = pd.DataFrame(classification_report(labels,predictions,output_dict = True)).round(decimals=2)
        
    return df


def load_occs(female=False,male=False):
    """Load occupations

    Keyword Arguments:
        female {bool} -- [Whether to include occupations that defined as stereotypically female] (default: {False})
        male {bool} -- [Whether to include occupations that defined as stereotypically male] (default: {False})

    Returns:
        [Two lists] -- [One list with all occupations, one list with occupations excluding the possessive case]
    """    
    all_occupations, occupations_no_poss, occupations_poss = [], [], []
    
    # could be loaded from file
    if male:
        occupations_male = ['chaufføren', 'supervisoren', 'viceværten', 'kokken', 'flyttemanden', 
        'bygningsarbejderen', 'lederen', 'udvikleren', 'tømreren', 'bestyreren', 
        'advokaten', 'landmanden', 'sælgeren', 'lægen', 'vagten', 
        'analytikeren', 'mekanikeren', 'direktøren']
        occupations_male_poss = [occ + 's' for occ in occupations_male] # possessive case

    # could be loaded from file
    if female:
        occupations_female = ['ekspedienten', 'læreren','sygeplejersken','assistenten','sekretæren',
        'revisoren','rengøringsassistenten','receptionisten','kontorassistenten','rådgiveren',
        'designeren','frisøren','skribenten','husholdersken','bageren',
        'bogholderen','redaktøren','bibliotekaren','syersken']
        occupations_female_poss = [occ + 's' for occ in occupations_female] # possessive case

    if male and not female: 
        # occ's without pos and with pos
        occupations_no_poss, occupations_poss = occupations_male, occupations_male_poss

    if female and not male: 
        # occ's without pos and with pos
        occupations_no_poss, occupations_poss = occupations_female, occupations_female_poss

    if male and female:
        # list with occ's without possessive case
        occupations_no_poss = occupations_male + occupations_female
        # list with occ's with possessive case
        occupations_poss = occupations_male_poss + occupations_female_poss

    # list with all occupations
    all_occupations = occupations_no_poss + occupations_poss

    return all_occupations, occupations_no_poss

def run_winobias_coref(coref_model, nlp):
    """Run winobias coref experiment
    Args:
        coref_model_name: name of coref model
    Returns:
        None
    """
    
    # data paths
    inpath_pro = os.path.join(Path(__file__).parents[1],"data","DaWinoBias_pro_stereotyped_evalda.txt")
    inpath_anti = os.path.join(Path(__file__).parents[1],"data","DaWinoBias_anti_stereotyped_evalda.txt")

    # load data - nb astrid updates this fn 
    anti_lines = load_texts(inpath_anti)
    pro_lines = load_texts(inpath_pro)

    #define occupations
    occupations_male, _ = load_occs(male=True)
    occupations_female, _ = load_occs(female=True)

    occupations = occupations_male + occupations_female

    # go through each condition
    for condition in ['anti_stereotypical', 'pro_stereotypical']:
        #print(f"Condition: {condition}")
        if condition == 'anti_stereotypical':  
            lines = anti_lines 
        elif condition == 'pro_stereotypical':  
            lines = pro_lines

        # prediction results: [successful preds, unsuccessful preds, failed preds]
        pred_res = [0,0,0]
        labels, preds = [], []
        labels_occ, preds_occ = [], []
        labels_stereotype, preds_stereotype = [], []
        
        #PROGESS BAR
        bar = progressbar.ProgressBar(maxval=len(lines)).start()

        # Look through sentences
        for idx, line in enumerate(lines): 
            # read line as spacy doc
            line = nlp(line)

            # tokenize and lowercase
            tokens = []
            for token in line:
                tokens.append(token.text.lower())

            # get correct coref and incorrect coref to compare with predictions
            coref_res,_ = idx_occ_pron(tokens)
            
            # remove square brackets aka the solution before feeding into coref model
            tokens = remove_sq_br(tokens)
            
            # apply coreference resolution to the document and get a list of predicted coref clusters (not idx)
            clusters = coref_model.predict_clusters(tokens)

            # get token indices from predicted cluster
            
            #if len(clusters) == 0:
            if clusters == []:
                cluster_idx = [-1]
            else:
                cluster_idx = [i[1] for i in clusters[0]]
            
            # compare predicted clusters with correct res
            if cluster_idx == coref_res[0]:
                pred_res[0] += 1
            elif cluster_idx == coref_res[1]:
                pred_res[1] += 1
            else: 
                pred_res[2] += 1

            # labels 
            labels.append(coref_res[0][0])
            
            #predictions
            if len(cluster_idx)>2 or len(cluster_idx)<1: 
                preds.append(-1)
            elif len(cluster_idx)==2:
                preds.append(cluster_idx[0])
            
            #occupation labels
            labels_occ.append(tokens[0][coref_res[0][0]])
            preds_occ.append(tokens[0][cluster_idx[0]])

            #remove possisive occupations 
            labels_occ = [remove_suffix(label, 's') for label in labels_occ]
            preds_occ = [remove_suffix(pred, 's') for pred in preds_occ]

            #group occupations
            for label, pred in zip(labels_occ, preds_occ):
                if label in occupations_female and pred in occupations: 
                    labels_stereotype.append('stereotypical female')
                elif label in occupations_male  and pred in occupations:
                    labels_stereotype.append('stereotypical male')

            for pred in preds_occ:
                if pred in occupations_female: 
                    preds_stereotype.append('stereotypical female')
                elif pred in occupations_male:
                    preds_stereotype.append('stereotypical male')

            # update bar
            bar.update(idx)
        bar.finish()

        #remove invalid predictions
        labels_occ = [labels_occ for labels_occ, preds_occ in zip(labels_occ, preds_occ) if preds_occ in occupations]
        preds_occ = [preds_occ for preds_occ in preds_occ if preds_occ in occupations]

        # get results in table
        # results per occupations, save file
        evaluate_model(labels_occ, preds_occ) #, filename = f'output/winobias_{coref_model_name}_{condition}_occupations') #switch to clf function
        # results per stereotype, save file
        wino_res_ = evaluate_model(labels_stereotype, preds_stereotype) #, filename = f'output/wino_bias{coref_model_name}_{condition}_stereotypes') #switch to clf function

        if condition == 'anti_stereotypical':
            anti_res = wino_res_
        elif condition == 'pro_stereotypical':
            pro_res = wino_res_
    return anti_res, pro_res

def evaluate_coref_winobias(anti_res, pro_res, model_name):
    """Evaluate winobias coref experiment
    Args:
        anti_res: results for anti-stereotypical condition
        pro_res: results for pro-stereotypical condition
        
    """
    # Gender Effect Size calculation
    anti_acc = anti_res["accuracy"].values[0]
    pro_acc = pro_res["accuracy"].values[0]
    gender_effect_size = logratio(pro_acc, anti_acc)

    anti_f1_fem_pron = anti_res["stereotypical female"].values[2]
    anti_f1_male_pron = anti_res["stereotypical male"].values[2]

    pro_f1_fem_pron = pro_res["stereotypical female"].values[2]
    pro_f1_male_pron = pro_res["stereotypical male"].values[2]

    simlpe_data = {
        f'Simple Output for {model_name}': ['', 'DaWinoBias'],
        'Gender Effect Size': ['', gender_effect_size],
        'Condition': ['Anti-stereotypical','Pro-stereotypical'],
        'Accuracy': [anti_acc, pro_acc]
        }

    simlpe_data = pd.DataFrame(simlpe_data).T

    # Set the first row as the column names
    simlpe_data.columns = simlpe_data.iloc[0]

    # Drop the first row
    simple_df = simlpe_data.iloc[1:]

    detailed_data = {
        f'Detailed Output for {model_name}': ['', 'DaWinoBias', '',''],
        'Gender Effect Size': ['', gender_effect_size, '',''],
        'Condition': ['Anti-stereotypical','', 'Pro-stereotypical', ''],
        'Accuracy': [anti_acc, '', pro_acc, ''],
        'Pronouns': ['Female', 'Male', 'Female', 'Male'],
        'F1': [anti_f1_fem_pron, anti_f1_male_pron, pro_f1_fem_pron, pro_f1_male_pron]
        }

    detailed_data = pd.DataFrame(detailed_data).T

    # Set the first row as the column names
    detailed_data.columns = detailed_data.iloc[0]

    # Drop the first row
    detailed_df = detailed_data.iloc[1:]

    results = [simple_df, detailed_df]

    return results




"""
Utility functions for running the ABC coreference task.

Builds on the codebase developed for the following paper: 

Title: “Type B reflexivization as an unambiguous testbed for multilingual multi-task gender bias.”
Authors: González, A. V., Barrett, M., Hvingelby, R., Webster, K., & Søgaard, A.
Date: 2020
Code availability: https://github.com/anavaleriagonzalez/ABC-dataset
"""
import json
import math
import os
import random
import sys
from pathlib import Path

import fairlearn.metrics as flm
import numpy as np
import pandas as pd
import spacy
import torch
from sklearn.exceptions import UndefinedMetricWarning
from sklearn.metrics import classification_report, f1_score


def run_abc_coref(coref_model):
    ''' 
    Run the coreference model on the ABC dataset and return predictions. (Duration approx. 20 min.)
    '''
    # load stereotypically male occupation sentences
    with open(os.path.join(Path(__file__).parents[1], "data", "abc_male_sents.txt"), "r") as f:
        male_abc = f.readlines()
    # load stereotypically female occupation sentences
    with open(os.path.join(Path(__file__).parents[1], "data", "abc_fem_sents.txt"), "r") as f:
        fem_abc = f.readlines()

    all_preds = []
    for gendered_data in [fem_abc, male_abc]:  
        data_ = [line_ for line_ in gendered_data if line_ != '---\n']
        i = 0
        preds = []

        #PROGRESS BAR
        abc_bar = progressbar.ProgressBar(maxval=len(data_)).start()
        
        # run the model on the data with the predict method and save the predictions
        for idx, i in enumerate(data_): 
            line = [i.split()]
            try:
            # get preds  
                predicted_clusters = coref_model.predict(line)
            except: 
                print(line)
            preds.append(predicted_clusters)
            abc_bar.update(idx)
        abc_bar.finish()
        
        all_preds.append(preds)
    fem_preds = all_preds[0]
    male_preds = all_preds[1]

    return fem_preds, male_preds
        
def chunks(lst, n):
    """Yield successive n-sized chunks from lst."""
    for i in range(0, len(lst), n):
        yield lst[i:i + n]

def get_coref_predictions(chunk_preds):
    """Get the coreference predictions for the reflexive case and the anti-reflexive cases.
    
    Args:
        - chunk_preds: list of predictions
        - coref_output_file: the name of the file where the predictions are stored
    
    Returns:
        - fem
        - male
        - reflexive
    """
    
    ref, male, fem = [], [], []
    
    # the truth labels for the reflexive case, i.e. all 1s
    labels_ref = list(np.repeat(1, len(chunk_preds)))
    # the truth labels for the anti-reflexive cases, i.e. all 0s
    labels_anti_ref = list(np.repeat(0, len(chunk_preds)))
    
    for p in chunk_preds:
        #assert if %3 == 0
        
        ref_pred, male_pred,fem_pred = p[0], p[1], p[2]
        
        predicted_clusters = "clusters" # could be an input argument to the function to ensure that the correct key is used in case of tesitng pther models

        cluster_ref = ref_pred[predicted_clusters] 
        cluster_male = male_pred[predicted_clusters]
        cluster_fem = fem_pred[predicted_clusters] 
        clusters = [cluster_ref, cluster_male, cluster_fem]   

        for i, cluster in enumerate(clusters):
            if i == 0:
                if cluster != []:
                    ref.append(1)
                else:
                    ref.append(0)
            elif i==1:
                if cluster != []:
                    male.append(1)
                else:
                    male.append(0)
            elif i==2:
                if cluster != []:
                    fem.append(1)
                else:
                    fem.append(0)
    return ref, fem, male, labels_ref, labels_anti_ref

def get_positive_preds(chunk_preds):
    '''Get False Posivie Rates for the reflexive case and the anti-reflexive cases and print output tables.'''
    
    ref, fem, male, labels_ref, labels_anti_ref = get_coref_predictions(chunk_preds)
    labels_preds_ = zip([labels_ref,labels_anti_ref,labels_anti_ref], [ref,fem,male])                
    # the tags for the classification reports
    tags = ["reflexive", "anti_reflexive_female", "anti_reflexive_male"] 
    res = {}
    for idx, (label, pred) in enumerate(labels_preds_):
        
        if label == labels_ref:
            # get the true positive rate for the current label and prediction
            res[f"tpr_{tags[idx]}"] = flm.true_positive_rate(label, pred, pos_label=1)
        
        elif label == labels_anti_ref:
            # get the false positive rate for the current label and prediction
            res[f"fpr_{tags[idx]}"] = flm.false_positive_rate(label, pred, pos_label=1)
    print(f"""
    Total positive prediction of coreference clusters: 
    
    Female: {np.sum(fem)}
    Male: {np.sum(male)}
    Reflexive: {np.sum(ref)}
    Total predictions:  {len(fem)} {len(male)} {len(ref)}""")

    return res

def evaluate_coref_abc(fem_preds, male_preds): 
    
    for gender in ["fem", "male", "all_sents"]: #and for all 
        
        if gender == "fem":
            preds = fem_preds
        elif gender == "male":
            preds = male_preds
        elif gender == "all_sents":
            preds = fem_preds +male_preds # merge the two lists
        
        # split the predictions into chunks of 3
        chunk_preds = list(chunks(preds, 3)) #danish

        # get the false positive rates for the reflexive and anti-reflexive cases
        results = get_positive_preds(chunk_preds)

        # save the results in a dataframe
        if gender == "fem":
            df_fem = results
        elif gender == "male":
            df_male = results
        elif gender == "all_sents":
            df_all_sents = results
    
    return df_fem, df_male, df_all_sents #dicts 

def logratio(x:float ,y:float):
    return round(math.log2(x/ y), 2)

def eval_results(fem:dict, male:dict, all:dict, model_name:str):
    '''Evaluate the results of the coreference model on the ABC data.'''    
    
    data = [fem['tpr_reflexive'],
           fem['fpr_anti_reflexive_female'], 
           fem['fpr_anti_reflexive_male'], 

           male['tpr_reflexive'], 
           male['fpr_anti_reflexive_female'],
           male['fpr_anti_reflexive_male'],
           
           all['tpr_reflexive'],
           all['fpr_anti_reflexive_female'],
           all['fpr_anti_reflexive_male']
           ]
    
    mean_refl = np.mean([fem['tpr_reflexive'], male['tpr_reflexive']])
    mean_fem = np.mean([fem['fpr_anti_reflexive_female'], male['fpr_anti_reflexive_female']])
    mean_male = np.mean([fem['fpr_anti_reflexive_male'], male['fpr_anti_reflexive_male']])
    
    gender_effect_size = logratio(mean_male, mean_fem)

    data = [[round(data_,2) for data_ in data]]
    
    simlpe_data = {
        f'Simple Output for {model_name}': ['', 'ABC', ''],
        'Gender Effect Size': ['', gender_effect_size, ''],
        'Pronoun': ['Female', 'Male', 'Reflexive'],
        'Mean Rate of Detected Clusters': [mean_fem, mean_male, mean_refl]
    }

    simlpe_data = pd.DataFrame(simlpe_data).T

    # Set the first row as the column names
    simlpe_data.columns = simlpe_data.iloc[0]

    # Drop the first row
    simple_df = simlpe_data.iloc[1:]

    detailed_data = {
        f'Detailed Output for {model_name}': ['', '', 'ABC', '', '', '' ],
        'Gender Effect Size': ['', '', gender_effect_size, '', '', ''],
        'Pronoun': ['', 'Female', '', 'Male', '', 'Reflexive'],
        'Mean Rate of Detected Clusters': ['', mean_fem, '', mean_male, '', mean_refl],
        'Stereotypical Occupation': ['Female', 'Male', 'Female', 'Male', 'Female', 'Male'],
        'Rate of Detected Clusters': [fem['fpr_anti_reflexive_female'], male['fpr_anti_reflexive_female'], fem['fpr_anti_reflexive_male'], male['fpr_anti_reflexive_male'], fem['tpr_reflexive'], male['tpr_reflexive']]
        
    }

    detailed_data = pd.DataFrame(detailed_data).T

    # Set the first row as the column names
    detailed_data.columns = detailed_data.iloc[0]

    # Drop the first row
    detailed_df = detailed_data.iloc[1:]
    results = [simple_df, detailed_df]

    return results


#from transformers import logging
#logging.set_verbosity_error()
#warnings.simplefilter("ignore", UserWarning)

import warnings
warnings.simplefilter("ignore", FutureWarning)

class Evaluator:
    """Module for detecting gender bias in Danish language models."""
    def __init__(self, model_name):
        self.model_name = model_name
        print(f'[INFO] You can test {self.model_name} by running Evaluator.evalute_<model type>()')
    
    def evaluate_ner(self, n):
        """Evaluate gender bias in a NER model.
        This function can be used for running the DaNe dataset test. 
        Read more about the specifics of these test in section ??? in the docs.   
        
        Args:
            n (int): Number of repetitions to run the augmentation pipeline. To ensure robustness we recommend a value of n => 20.

        Returns:
            list: Performance output as list. First element: performance in condensed form. Second element: performance in detailed form.
        """        
        from dacy.datasets import dane

        from .ner_tasks.augmentation import (f_aug, m_aug,
                                                        muslim_f_aug,
                                                        muslim_m_aug)
        from .ner_tasks.performance import load_mdl, eval_model_augmentation
        testdata = dane(splits=["test"], redownload=True, open_unverified_connected=True)

        model = load_mdl(self.model_name)
        if n <= 1:
            print(f"[INFO] Please choose a value for n larger than 1 to ensure robustness, got: {n}.")
            print(f"[INFO] Running the NER task on {self.model_name} with low value for n.")
        else:
            print(f"[INFO] Running the NER task on {self.model_name}")
        
        # define augmenters 
        augmenters = [
            (f_aug, "Majority female names", n),
            (m_aug, "Majority male names", n),
            (muslim_f_aug, "Minority female names", n),
            (muslim_m_aug, "Minority male names", n)]

        # run model 
        output = eval_model_augmentation(model, self.model_name, str(n), augmenters, testdata)
        print("[INFO] output(list) generated. Access condensed output with output[0], and detailed output with with output[1].")
        return output
    
    def evaluate_pretrained(self, test, mask_token=None, start_token = None, sep_token=None):
        """Evaluate gender bias in a pre-trained model trained with masked language modeling. 

        This function can be used for running two different tests:
        The Dawinobias Language Modeling Task and the ABC Language Modeling Task. 
        Read more about the specifics of these test in section ??? in the docs.   

        Args:
            test (str): choose between "abc" or "dawinobias"
            mask_token (str, optional): mask token of tested model. Specify when running test "abc". Defaults to None.
            start_token (str, optional): start token of tested model. Specify when running test "abc". Defaults to None.
            sep_token (str, optional): sep token of tested model. Specify when running test "dawinobias". Defaults to None.

        Raises:
            ValueError: If no start_token is specified when running the ABC test. 
            ValueError: If no sep_token is specified when running the ABC test. 
            ValueError: If no mask_token is specified when running the dawinobias test. 
            ValueError: If non-existing task is specified.  

        Returns:
            list: Performance output as list. First element: performance in condensed form. Second element: performance in detailed form.
        """        
        import pandas as pd
        import spacy
        from transformers import pipeline

        from .lm_tasks.abc_utils import get_output, load_abc, load_mdl, run_abc
        from .lm_tasks.wino_utils import evaluate_lm_winobias, run_winobias

        ### RUN ABC
        # load data 
        if test == "abc":
            if start_token is None:
                raise ValueError("Please specify input argument 'start_token'(str) when running the ABC language modeling task.")
            if sep_token is None:
                raise ValueError("Please specify input argument 'sep_token'(str) when running the ABC language modeling task.")
            else:
                pass
            
            print(f"[INFO] Running the ABC language modeling task on {self.model_name}")
            refl_sents_m, refl_sents_f = load_abc()
            # load tokenizer and model
            model, tokenizer = load_mdl(self.model_name)
            
            # create results df 
            out_df_f = run_abc(refl_sents_f, "female", tokenizer, model, start_token, sep_token)
            out_df_m = run_abc(refl_sents_m, "male", tokenizer, model, start_token, sep_token)


            # evaluate abc
            results = get_output(out_df_f, out_df_m, model_name = self.model_name)

        elif test == "dawinobias":
            if mask_token is None:
                raise ValueError("Please specify input argument 'mask_token'(str) when running the DaWinobias language modeling task.")
            else:
                pass
            print(f"[INFO] Running the DaWinobias language modeling task on {self.model_name}")            
            #load model used for tokenization
            try:
                tokenizer = spacy.load("da_core_news_sm")
            except OSError: 
                print("[INFO] Downloading tokenizer: da_core_news_sm from spaCy.")
                from spacy.cli.download import download
                download("da_core_news_sm")
                tokenizer = spacy.load("da_core_news_sm")
            
            # initiate pipeline
            print(f"[INFO] Loading model {self.model_name} from Hugging Face.")
            nlp =  pipeline(task = "fill-mask", model = self.model_name) 

            # run wino
            clf_rep_anti, clf_rep_pro= run_winobias(tokenizer, nlp, mask_token = mask_token, model_name = self.model_name)
            
            results = evaluate_lm_winobias(clf_rep_anti, clf_rep_pro, model_name = self.model_name)
                    
        else: 
            raise ValueError("Not a valid test. Choose between 'abc' and 'dawinobias'")
        print("[INFO] output(list) generated. Access condensed output with output[0], and detailed output with with output[1].")
        return results 
        
    def evaluate_coref(self, test, model):
        """Evaluate gender bias in a coreference model. 

        This function can be used for running two different tests:
        The Dawinobias Language Coreference Task and the ABC Coreference Task. 
        Read more about the specifics of these test in section ??? in the docs.   

        Args:
            test (str): choose between "abc" or "dawinobias"
            model (_type_): a coreference model object. 
        Raises:
            ValueError: If non-existing task is specified.  

        Returns:
            list: Performance output as list. First element: performance in condensed form. Second element: performance in detailed form.
        """              

        #from danlp.models import load_xlmr_coref_model
        import os
        import random
        import sys
        #import json
        from pathlib import Path

        import numpy as np
        import pandas as pd
        import spacy
        import torch
        from sklearn.metrics import classification_report, f1_score

        # load the coreference model
        #if model_name == "xlm-r_coref": 
            #print("Loading the XLM-R coreference model")
            #model = load_xlmr_coref_model() 

        if test == "dawinobias":
            import nltk
            from .coref_tasks.wino_utils import (evaluate_coref_winobias,
                                                  run_winobias_coref)
            nltk.download('omw-1.4') #only wino
            #load model used for tokenization
            try:
                nlp = spacy.load("da_core_news_sm")
            except OSError: 
                print("[INFO] Downloading tokenizer: da_core_news_sm from spaCy.")
                from spacy.cli.download import download
                download("da_core_news_sm")
                nlp = spacy.load("da_core_news_sm")

            print(f"[INFO] Running the DaWinobias coreference task on {self.model_name}")
            anti_res, pro_res =run_winobias_coref(model, nlp)
            results = evaluate_coref_winobias(anti_res, pro_res, model_name = self.model_name)
            
        elif test == "abc":
            import pandas as pd
            from .coref_tasks.abc_utils import (eval_results,
                                                  evaluate_coref_abc,
                                                  run_abc_coref)
            print(f"[INFO] Running the ABC coreference task on {self.model_name}")
            fem_preds, male_preds = run_abc_coref(model)
            #two dicts of f1 scores
            df_fem, df_male, all_sents = evaluate_coref_abc(fem_preds= fem_preds, male_preds= male_preds)

            results = eval_results(df_fem, df_male, all_sents, model_name=self.model_name)
        else:
            raise ValueError("Not a valid test. Choose between 'abc' and 'dawinobias'")
        print("[INFO] output(list) generated. Access condensed output with output[0], and detailed output with with output[1].")
        return results 

class Visualizer:
    """Module for visualizing the results of the evaluation"""
    def __init__(self):
        import seaborn as sns
        sns.set_style("whitegrid", rc={"lines.linewidth": 1})
        sns.set_context("notebook", font_scale=1.2)

        #self.palette = sns.color_palette("pastel",n_colors=15).reverse()
    
    def visualize_results(self, df, framework, model_name, task=None):
        """Visualize output from any of the genderbias tests that can be run in this package. 

        Args:
            df (df): detailed output from any of the tests. 
            framework (str): choose between NER, Dawinobias or ABC. 
            model_name (str): model name
            task (str, optional): For coref or pretrained model specify whether the output is from the abc test or the dawinobias test. Defaults to None.

        Returns:
            plot: seaborn plot visualization. 
        """        
        import matplotlib.pyplot as plt
        import pandas as pd
        import seaborn as sns

        sns.set_style("whitegrid", rc={"lines.linewidth": 10})
        
        df = pd.DataFrame()

        if framework == "abc":
            df["Stereotypical Occupations"] = ["Female occupations","Male occupations", "Main Effect"] *2
            df["Anti-reflexive Pronoun"] = ["Female"] *3 + ["Male"] *3
            markers = ["o", "o", "o"]
            x = "Anti-reflexive Pronoun"
            nuance = "Stereotypical Occupations"

            if task == "coref":
                try:
                    mean_fem = list(df.loc["Mean Rate of Detected Clusters"])[1]
                    mean_male = list(df.loc["Mean Rate of Detected Clusters"])[3]
                    fpr_fem_pron_fem_occ = list(df.loc["Rate of Detected Clusters"])[0]
                    fpr_fem_pron_male_occ = list(df.loc["Rate of Detected Clusters"])[1]
                    fpr_male_pron_fem_occ = list(df.loc["Rate of Detected Clusters"])[2]
                    fpr_male_pron_male_occ = list(df.loc["Rate of Detected Clusters"])[3]
                except:
                    mean_fem = float(df.iloc[2,2])
                    mean_male = float(df.iloc[2,4])
                    fpr_fem_pron_fem_occ = float(df.iloc[4,1])
                    fpr_fem_pron_male_occ = float(df.iloc[4,2])
                    fpr_male_pron_fem_occ = float(df.iloc[4,3])
                    fpr_male_pron_male_occ = float(df.iloc[4,4])
                
                points = [  fpr_fem_pron_fem_occ,
                            fpr_fem_pron_male_occ,
                            mean_fem,
                            fpr_male_pron_fem_occ,
                            fpr_male_pron_male_occ,
                            mean_male] 

                df["False Positive Rates"] = points
                y = "False Positive Rates"
                title = f"ABC Coref Task: {model_name}"
            
            elif task == "lm":
                try:
                    points = [float(df.iloc[4,1].split(" ")[0]), 
                        float(df.iloc[4,2].split(" ")[0]),
                        float(df.iloc[2,1].split(" ")[0]),
                        float(df.iloc[4,3].split(" ")[0]), 
                        float(df.iloc[4,4].split(" ")[0]),
                        float(df.iloc[2,3].split(" ")[0])]
                except:
                    points = [float(df.iloc[4,0].split(" ")[0]), 
                        float(df.iloc[4,1].split(" ")[0]),
                        float(df.iloc[2,0].split(" ")[0]),
                        float(df.iloc[4,2].split(" ")[0]), 
                        float(df.iloc[4,3].split(" ")[0]),
                        float(df.iloc[2,2].split(" ")[0])]
                df["Median Perplexity"] = points
                y = "Median Perplexity"
                title = f"ABC LM Task: {model_name}"
        
        elif framework == "dawinobias":
            df["Pronoun"] = ["Female (F1)","Male (F1)", "Main Effect (Accuracy)"]*2
            if task == "coref":
                try:
                    points = [float(df.iloc[4,1]), 
                        float(df.iloc[4,2]), 
                        float(df.iloc[2,1]), 
                        float(df.iloc[4,3]), 
                        float(df.iloc[4,4]), 
                        float(df.iloc[2,3])]
                except:
                    points =  [df.loc["F1"][0],
                        df.loc["F1"][1],
                        df.loc["Accuracy"][0],
                        df.loc["F1"][2],
                        df.loc["F1"][3],
                        df.loc["Accuracy"][2]]
                title = f"DaWinoBias, Coreference Task: {model_name}"
            elif task == "lm":
                try: #if loaded
                    points = [
                        float(df.iloc[4,1]),
                        float(df.iloc[4,2]),
                        float(df.iloc[2,1]),
                        float(df.iloc[4,3]),
                        float(df.iloc[4,4]),
                        float(df.iloc[2,3])
                    ]
                except: # if 
                    points =  [df.loc["F1"][0],
                        df.loc["F1"][1],
                        df.loc["Accuracy"][0],
                        df.loc["F1"][2],
                        df.loc["F1"][3],
                        df.loc["Accuracy"][2]]
                title = f"DaWinoBias, LM Task: {model_name}"

            df["Performance"] = points
            df["Condition"] = ["Anti-stereotypical"]*3 + ["Pro-stereotypical"]*3
            x = "Condition"
            y = "Performance"
            nuance = "Pronoun"
            markers = ["o", "o", "o"]
            title = title

        elif framework == "ner":
            df["Protected Group"] = ["Majority (F1)", "Minority (F1)"]*2
            try:    
                points = [float(df.iloc[2,1].split(" ")[0]), 
                    float(df.iloc[4,2].split(" ")[0]), 
                    float(df.iloc[2,3].split(" ")[0]), 
                    float(df.iloc[4,4].split(" ")[0])]
            except:
                points = [float(df.iloc[4,0].split(" ")[0]),
                        float(df.iloc[4,1].split(" ")[0]), 
                        float(df.iloc[4,2].split(" ")[0]), 
                        float(df.iloc[4,3].split(" ")[0])]

            df["Performance"] = points
            df["Augmentation"] = ["Female Names"] *2 + ["Male Names"]*2
            x = "Augmentation"
            y = "Performance"
            nuance = "Protected Group"
            markers = ["o", "o"]
            title = f"NER Task, Augmented DaNe: {model_name}"

        sns.pointplot(data=df, 
                    x=x,
                    y=y, 
                    hue= nuance,  
                    dodge=False, 
                    join=True, 
                    markers= markers, 
                    scale=1.2, 
                    linestyles=[':', ':', '-'],
                    palette=["sandybrown", "mediumpurple", "darkgrey"]).set_title(title)
        
        plt.minorticks_on()

        return plt


def load_mdl(model_name):
    from transformers import AutoModelForMaskedLM, AutoTokenizer
    model = AutoModelForMaskedLM.from_pretrained(model_name)
    tokenizer = AutoTokenizer.from_pretrained(model_name)
    model.eval()

    return model, tokenizer
